/* Message list header manipulation.
   Copyright (C) 2007, 2016-2017 Free Software Foundation, Inc.
   Written by Bruno Haible <bruno@clisp.org>, 2007.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */


#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

/* Specification.  */
#include "msgl-header.h"

#include <string.h>

#include "xalloc.h"
#include "xerror.h"
#include "xvasprintf.h"
#include "gettext.h"

#define SIZEOF(a) (sizeof(a) / sizeof(a[0]))
#define _(str) gettext (str)

/* The known fields in their usual order.  */
static const struct
{
  const char *name;
  size_t len;
}
known_fields[] =
  {
    { "Project-Id-Version:", sizeof ("Project-Id-Version:") - 1 },
    { "Report-Msgid-Bugs-To:", sizeof ("Report-Msgid-Bugs-To:") - 1 },
    { "POT-Creation-Date:", sizeof ("POT-Creation-Date:") - 1 },
    { "PO-Revision-Date:", sizeof ("PO-Revision-Date:") - 1 },
    { "Last-Translator:", sizeof ("Last-Translator:") - 1 },
    { "Language-Team:", sizeof ("Language-Team:") - 1 },
    { "Language:", sizeof ("Language:") - 1 },
    { "MIME-Version:", sizeof ("MIME-Version:") - 1 },
    { "Content-Type:", sizeof ("Content-Type:") - 1 },
    { "Content-Transfer-Encoding:", sizeof ("Content-Transfer-Encoding:") - 1 }
  };

void
msgdomain_lists_merge_headers (msgdomain_list_ty **mdlps,
				size_t nfiles)
{
  message_list_list_ty * headers = message_list_list_alloc ();
  char * plur = "Plural-Forms:";
  char * plurals[nfiles];
  int plur_index = 0;
  /* First, find all header entries and cut them to lines. */
  for (int n = 0; n < nfiles; n++)
    {
      msgdomain_list_ty *mdlp = mdlps[n];
      for (size_t k = 0; k < mdlp->nitems; k++)
        {
          message_list_ty * chopped = message_list_header_list (mdlp->item[k]->messages);
          if (chopped) message_list_list_append (headers, chopped);
        }

      /* Set plural to NULL by default. */
      plurals[n] = NULL;
    }
  /* While there are some messages remaining, take the first one. */
  while (headers->nitems > 0)
    {
      message_ty * field = headers->item[0]->item[0];
      /* Let us save plurals for later use. */
      if (strcmp(field->msgid, plur) == 0)
        {
          plurals[0] = XNMALLOC (field->msgstr_len+1, char);
          strcpy (plurals[0], field->msgstr);
          for (int i = 1; i < headers->nitems; i++)
            {
              message_ty * mp = message_list_search (headers->item[i], NULL, plur);
              if (mp!=NULL)
                {
                  plurals[i] = XNMALLOC (mp->msgstr_len+1, char);
                  strcpy (plurals[i], mp->msgstr);
                }
            }
        }
      /* Set the header field and delete all the occurences of the field. */
      msgdomain_list_set_header_field (mdlps[0], field->msgid, field->msgstr);
      for (int i = 1; i < headers->nitems; i++)
        {
          message_ty * mp = message_list_search (headers->item[i], NULL, field->msgid);
          if (mp != NULL)
            {
              /* If needed, fix line numbering in advance. */
              if (mp != headers->item[i]->item[0])
                for (int l = mp->pos.line_number - headers->item[i]->item[0]->pos.line_number + 1; l < headers->item[i]->nitems; l++)
                  headers->item[i]->item[l]->pos.line_number--;
              message_list_delete_nth (headers->item[i], mp->pos.line_number - headers->item[i]->item[0]->pos.line_number);
            }
        }
      message_list_delete_nth (headers->item[0], 0);
      /* If the first header is empty, start to process next nonempty header. */
      while (headers->nitems > 0 && headers->item[0]->nitems == 0)
        {
          message_list_free (headers->item[0], 0);
          for (int i = 0; i < headers->nitems - 1; i++)
            headers->item[i] = headers->item[i+1];
          headers->nitems--;
        }
    }

  /* Some plural manipulation. */
  char *res = NULL;
  char *prevres = NULL;
  prevres = plurals[0];
  /* The prevres is the value currently in the output header,
   * res is the value just read. So if res == NULL we just
   * continue, which is correct. */
  for (int n = 1; n < nfiles; n++)
    {
      res = plurals[n];
      if (res != NULL)
        {
          if (prevres == NULL)
            {
              msgdomain_list_set_header_field (mdlps[0], plur, res);
              prevres = res;
            }
          else if (strcmp (res, prevres) != 0)
            {
              multiline_error (xstrdup (""),
                               xasprintf (_("\
Input po files have different Plural-Forms. Invalid output file was created. \n\
Please, fix the plurals.\n")));
            }
        }
    }
}

void
msgdomain_list_set_header_field (msgdomain_list_ty *mdlp,
                                 const char *field, const char *value)
{
  size_t field_len;
  int field_index;
  size_t k, i;

  field_len = strlen (field);

  /* Search the field in known_fields[].  */
  field_index = -1;
  for (k = 0; k < SIZEOF (known_fields); k++)
    if (strcmp (known_fields[k].name, field) == 0)
      {
        field_index = k;
        break;
      }

  for (i = 0; i < mdlp->nitems; i++)
    {
      message_list_ty *mlp = mdlp->item[i]->messages;
      size_t j;

      /* Search the header entry.  */
      for (j = 0; j < mlp->nitems; j++)
        if (is_header (mlp->item[j]) && !mlp->item[j]->obsolete)
          {
            message_ty *mp = mlp->item[j];

            /* Modify the header entry (it does not have to be present). */
            const char *header = (mp->msgstr != NULL) ? mp->msgstr : "\0";
            char *new_header =
              XNMALLOC (strlen (header) + 1
                        + strlen (field) + 1 + strlen (value) + 1 + 1,
                        char);

            /* Test whether the field already occurs in the header entry.  */
            const char *h;

            for (h = header; *h != '\0'; )
              {
                if (strncmp (h, field, field_len) == 0)
                  break;
                h = strchr (h, '\n');
                if (h == NULL)
                  break;
                h++;
              }
            if (h != NULL && *h != '\0')
              {
                /* Replace the field.  */
                char *p = new_header;
                memcpy (p, header, h - header);
                p += h - header;
                p = stpcpy (p, field);
                p = stpcpy (stpcpy (stpcpy (p, " "), value), "\n");
                h = strchr (h, '\n');
                if (h != NULL)
                  {
                    h++;
                    stpcpy (p, h);
                  }
              }
            else if (field_index < 0)
              {
                /* An unknown field.  Append it at the end.  */
                char *p = new_header;
                p = stpcpy (p, header);
                if (p > new_header && p[-1] != '\n')
                  *p++ = '\n';
                p = stpcpy (p, field);
                stpcpy (stpcpy (stpcpy (p, " "), value), "\n");
              }
            else
              {
                /* Find the appropriate position for inserting the field.  */
                for (h = header; *h != '\0'; )
                  {
                    /* Test whether h starts with a field name whose index is
                       > field_index.  */
                    for (k = field_index + 1; k < SIZEOF (known_fields); k++)
                      if (strncmp (h, known_fields[k].name, known_fields[k].len)
                          == 0)
                        break;
                    if (k < SIZEOF (known_fields))
                      break;
                    h = strchr (h, '\n');
                    if (h == NULL)
                      break;
                    h++;
                  }
                if (h != NULL && *h != '\0')
                  {
                    /* Insert the field at position h.  */
                    char *p = new_header;
                    memcpy (p, header, h - header);
                    p += h - header;
                    p = stpcpy (p, field);
                    p = stpcpy (stpcpy (stpcpy (p, " "), value), "\n");
                    stpcpy (p, h);
                  }
                else
                  {
                    /* Append it at the end.  */
                    char *p = new_header;
                    p = stpcpy (p, header);
                    if (p > new_header && p[-1] != '\n')
                      *p++ = '\n';
                    p = stpcpy (p, field);
                    stpcpy (stpcpy (stpcpy (p, " "), value), "\n");
                  }
              }

            mp->msgstr = new_header;
            mp->msgstr_len = strlen (new_header) + 1;
          }
    }
}


void
message_list_delete_header_field (message_list_ty *mlp,
                                  const char *field)
{
  size_t field_len = strlen (field);
  size_t j;

  /* Search the header entry.  */
  for (j = 0; j < mlp->nitems; j++)
    if (is_header (mlp->item[j]) && !mlp->item[j]->obsolete)
      {
        message_ty *mp = mlp->item[j];

        /* Modify the header entry.  */
        const char *header = mp->msgstr;

        /* Test whether the field occurs in the header entry.  */
        const char *h;

        for (h = header; *h != '\0'; )
          {
            if (strncmp (h, field, field_len) == 0)
              break;
            h = strchr (h, '\n');
            if (h == NULL)
              break;
            h++;
          }
        if (h != NULL && *h != '\0')
          {
            /* Delete the field.  */
            char *new_header = XCALLOC (strlen (header) + 1, char);

            char *p = new_header;
            memcpy (p, header, h - header);
            p += h - header;
            h = strchr (h, '\n');
            if (h != NULL)
              {
                h++;
                strcpy (p, h);
              }
            else
              *p = '\0';

            mp->msgstr = new_header;
            mp->msgstr_len = strlen (new_header) + 1;
          }
      }
}

void
message_list_read_header_field (message_list_ty *mlp,
                                  const char *field, char **where_ptr)
{
  size_t field_len = strlen (field);
  size_t j;

  *where_ptr = NULL;

  /* Search the header entry.  */
  for (j = 0; j < mlp->nitems; j++)
    if (is_header (mlp->item[j]) && !mlp->item[j]->obsolete)
      {
        /* We found the correct message. */
        message_ty *mp = mlp->item[j];
        const char *header = mp->msgstr;

        /* Test whether the field occurs in the header entry.  */
        const char *h;

        for (h = header; *h != '\0'; )
          {
            if (strncmp (h, field, field_len) == 0)
              break;
            /* Jump by lines. */
            h = strchr (h, '\n');
            if (h == NULL)
              break;
            h++;
          }
        if (h != NULL && *h != '\0')
          {
            /* We found it and it is nonempty. Read the value of the field.  */
            h += field_len + 1;
            char *enh = strchr (h, '\n');
            if (enh != NULL && *enh != '\0')
	      {
	        *where_ptr = (char *)XNMALLOC (((enh - h) + 1), char);
                memcpy (*where_ptr, h, enh - h);
                /* Make the string null-terminated. */
                (*where_ptr)[enh-h] = '\0';
              }
          }
      }
}

message_list_ty *
message_list_header_list (message_list_ty *mlp)
{
  size_t j;

  /* Search the header entry.  */
  for (j = 0; j < mlp->nitems; j++)
    if (is_header (mlp->item[j]) && !mlp->item[j]->obsolete)
      {
        /* We found the correct message. */
        message_ty *mp = mlp->item[j];
        const char *h = mp->msgstr;
        message_list_ty * header = message_list_alloc (false);
        int ctr = 0;

        while (*h != '\0')
          {
            char *enh = strchr (h, ':');
            enh++;
            char * msgid = (char *)XNMALLOC (((enh - h) + 1), char);
            memcpy (msgid, h, enh - h);
            /* Make the string null-terminated. */
            (msgid)[enh-h] = '\0';
            h = enh + 1;

            enh = strchr (h, '\n');
            if (enh != NULL)
              {
                char * msgstr = (char *)XNMALLOC (((enh - h) + 1), char);
                memcpy (msgstr, h, enh - h);
                /* Make the string null-terminated. */
                msgstr[enh-h] = '\0';
                lex_pos_ty pos = {NULL, ctr++};
                message_list_append (header, message_alloc (NULL, msgid, NULL, msgstr, enh - h, &pos));
                h = enh + 1;
              }
            else return NULL;
          }
        return header;
      }
  return NULL;
}
